//-------------------------------------------------------------------//
//                                                                   //
//  RInfo Copyright (C) Glenn Rosenthal, 1998.  All rights reserved. //
//                                                                   //   
//  The KANJISTK.DAT file this program extracts information from was //
//  generated by KINFO.EXE from KANJIDIC compiled by Jim Breen, who  //
//  holds the copyright on it.                                       //
//                                                                   //
//  The radical lookup tables used were originally developed by      //
//  Michael Raine and Derc Yamasaki.                                 //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//                                                                   //
//  This program was developed to build kanji indexes that are used to 
//  dirve JWPce's radical lookup feature.  
//
//  This program is expecting to read the files KANJISRK.DAT generated
//  by the KINFO.EXE program (from Jim Breen's KANJIDIC).  RADKANJI.DAT
//  and RADKANJI.IDX, from JWP 1.31 (original information generated by
//  Michael Raine and Derc Yamaski.  Finally, the files STROKNJI.DAT and 
//  STROKNJI.IDX can be read, but have no use.  Their function (to 
//  index kanji by stroke count has been replaced by the KANJISRK.DAT
//  file.
//
//  The program takes no imput arguments at the moment, and produces 
//  two output fils.  The program must be run from the same location
//  as the input data files.
//
//  The oputput files are as follows:
//
//      RRADICAL.DAT -- Data file for kanji sorted by radicals contained
//                      within the kanji.
//
//      RSTROKE.DAT  -- Data file for kanji sorted by stroke count.
//      RADICAL.ECU  -- Radical data in EUC format.
//      STROKE.EUC   -- Stroke count data in EUC format.
//
//
//  The structure of the produced database is as follows:
//
//      <index> -- The index consists of a number of radical_info
//                 structures stored in a row.  For the RRADICAL.DAT
//                 file, there are 241 radicals, stored with zero being
//                 the first radical.  For the RSTROKE.DAT file there
//                 are 1-30 stroke entries (offset at zero).  Each 
//                 entry indicates the number number of kanji for that
//                 group, and a pointer to the kanji within the data 
//                 part of the file.  
//
//                 In the case of RRADICAL.DAT a number of false 
//                 radicals (or alternaltive rendrings) have been 
//                 inculded in the databse.  Each of these false 
//                 radicals is indicated by a count of zero.
//      <data>  -- Actuall data.  Access is by reading the index 
//                 then seaking to the data location and reading a 
//                 number of kanji.  The kanji are stored directly
//                 as JIS values.
//
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "rinfo.h"

//-------------------------------------------------------------------
//
//  Static defintions:
//
//#define READ_STROKE       // If defined causes the program to read STROKNJI.DAT &
                            //   STROKNJI.IDX, instead of using KANJISRK.DAT.

#define MAX_STROKE  30      // Maximum number of strokes
#define SIZE_KANJI  2000    // Maximum number of kanji per radical

//-------------------------------------------------------------------
//
//  Static data.
//
static data_file stroke_count;      // KANJISRK.DAT file
static data_file radical_index;     // RADKANJI.IDX file
static data_file radical_data;      // RADKANJI.DAT file
#ifdef READ_STROKE
static data_file stroke_index;      // STROKNJI.IDX file
static data_file stroke_data;       // STROKNJI.DAT file
#endif

//-------------------------------------------------------------------
//
//  Basic routines.
//

//
//  Determine the size of a file.
//
//      file   -- FILE pointer for file.
//
//      RETURN -- file size in bytes.
//
static int file_size (FILE *file) {
  int size;
  fseek (file,0,SEEK_END);
  size = ftell(file);
  fseek (file,0,SEEK_SET);
  return (size);
}

//-------------------------------------------------------------------
//
//  High level process routines.
//

//
//  This routine writes the radical information out in EUC form.  This is intened for 
//  debugging, but may be later used to actually edit the radical inforamtion.
//
//      name   -- Name of file to write.
//
//      RETURN -- A non-zero value inciates an error.
//
static int ecu_radical (char *name) {
  int          i,j,k;
  ushort       offset;
  KANJI        temp,kan,kanji[SIZE_KANJI];
  FILE        *file;
  radical_info info;
//
//  Intialize.
//
  printf ("Writing kanji radical information in EUC format [%s] ... ",name);
  if (!(file = fopen(name,"wb"))) { printf ("Unable to open file %s for writing\n",name); return (true); }
//
//  Write loop
//
  offset = radical_index.size*sizeof(radical_info);     // Intialize offset to past last radical
  for (i = 0; i < radical_index.size; i++) {
    fprintf (file,"%d\t",i+1);
    info.count  = radical_index.info[i].count;
    info.offset = offset;
    if (info.count > SIZE_KANJI) { printf ("\nSIZE_KANJI is exceeded, size=%d  radical=%d\n",info.count,i); fclose (file); exit (5); }
    memcpy (kanji,&radical_data.kanji[radical_index.info[i].offset/2],info.count*sizeof(KANJI));
    for (j = 0; j < info.count; j++) {                  // Sort kanji by stroke count.
      for (k = j+1; k < info.count; k++) {
        if (stroke_count.stroke[kanji[k]] < stroke_count.stroke[kanji[j]]) {
          temp     = kanji[j];
          kanji[j] = kanji[k];
          kanji[k] = temp;
        }
      }
    }
    for (j = 0; j < info.count; j++) {
      kan  = kanji[j];
      kan  = (kan >> 8) | ((kan & 0xff) << 8);
      kan |= 0x8080;
      fwrite (&kan,sizeof(kan),1,file);
    }
    fputs ("\n\n",file);
  }
//
//  Clean up.
//
  fclose (file);
  printf ("done\n");
  return (false);
}

//
//  This routine writes the stroke information out in EUC format.  This is primarally intended
//  as a debugging tool, but may later be used to edit the list, then this would become the 
//  input format.
//
static int euc_stroke (char *name) {
  int          i,j;
  KANJI        kanji;
  FILE        *file;
//
//  Introduction.
//
  printf ("Writing kanji stroke information int EUC format [%s] ... ",name);
  if (!(file = fopen(name,"wb"))) { printf ("Unable to open file %s for writing\n",name); return (true); }
//
//  Write block
//
  for (i = 0; i < MAX_STROKE; i++) {                   
    fprintf (file,"%d\t",i+1);
    for (j = 0; j < stroke_count.size; j++) {           // Count number of kanji.
      if (stroke_count.stroke[j] != i+1) continue;
      kanji  = (KANJI) j;
      kanji  = (kanji >> 8) | ((kanji & 0xff) << 8);
      kanji |= 0x8080;
      fwrite (&kanji,sizeof(kanji),1,file);             //   and write to disk while counting.
    }
    fputs ("\n\n",file);
  }
//
//  Clean up and exit.
//
  fclose (file);
  printf ("done\n");
  return (false);
}

//
//  Read an input file.
//
//      data   -- Structure to hold file being read.
//      name   -- Name of file to read.
//
//      RETURN -- A non-zero value indicates an error.
//
static int load_file (data_file *data,char *name) {
  FILE *file;
  printf ("Reading data file %s ... ",name);
  memset (data,0,sizeof(data_file));
  if (!(file = fopen(name,"rb"))) { printf ("Cannot open file %s\n",name); exit (5); }
  data->size = file_size(file);
  fread (data->kanji,data->size,1,file);
  fclose (file);
  printf ("done\n");
  return (false);
}

//
//  Write the radical database.
//
//      name -- name of file to write.
//
//      RETURN -- A non-zero value indicates an error.
//

static int write_radical (char *name) {
  int          i,j,k;
  ushort       offset;
  KANJI        temp,kanji[SIZE_KANJI];
  FILE        *file;
  radical_info info;
//
//  Intialize.
//
  printf ("Writing kanji radical information [%s] ... ",name);
  if (!(file = fopen(name,"wb"))) { printf ("Unable to open file %s for writing\n",name); return (true); }
//
//  Write loop
//
  offset = radical_index.size*sizeof(radical_info);     // Intialize offset to past last radical
  for (i = 0; i < radical_index.size; i++) {
    info.count  = radical_index.info[i].count;
    info.offset = offset;
    if (info.count > SIZE_KANJI) { printf ("\nSIZE_KANJI is exceeded, size=%d  radical=%d\n",info.count,i); fclose (file); exit (5); }
    memcpy (kanji,&radical_data.kanji[radical_index.info[i].offset/2],info.count*sizeof(KANJI));
    for (j = 0; j < info.count; j++) {                  // Sort kanji by stroke count.
      for (k = j+1; k < info.count; k++) {
        if (stroke_count.stroke[kanji[k]] < stroke_count.stroke[kanji[j]]) {
          temp     = kanji[j];
          kanji[j] = kanji[k];
          kanji[k] = temp;
        }
      }
    }
    fseek  (file,i*sizeof(radical_info),SEEK_SET);      // Write data to file
    fwrite (&info,sizeof(radical_info),1,file);
    fseek  (file,offset,SEEK_SET);
    fwrite (kanji,info.count,sizeof(KANJI),file);
    offset += info.count*sizeof(KANJI);
  }
//
//  Clean up.
//
  fclose (file);
  printf ("done [size=%d]\n",offset);
  return (false);
}

//
//  Write the stroke count data base.
//
//      name -- name of file to write.
//
//      RETURN -- A non-zero value indicates an error.
//
static int write_stroke  (char *name) {
  int          i,j,total;
  ushort       offset;
  KANJI        kanji;
  FILE        *file;
  radical_info info;
//
//  Introduction.
//
  printf ("Writing kanji stroke information [%s] ... ",name);
  if (!(file = fopen(name,"wb"))) { printf ("Unable to open file %s for writing\n",name); return (true); }
//
//  Make sure allocated stroke size (MAX_STROKE) is sufficient.
//
  for (i = 0; i < stroke_count.size; i++) {
    if (stroke_count.stroke[i] > MAX_STROKE) printf ("\nMAX_STROKE is exceded, stroke=%d  JIS=%x\n",stroke_count.stroke[i],i);
  }
//
//  Write block
//
  offset = MAX_STROKE*sizeof(radical_info);             // Intialize offset to the end of the index.
  total  = 0;                                           // Total of all kanji in list.
  for (i = 0; i < MAX_STROKE; i++) {                   
    fseek (file,offset,SEEK_SET);                       // Seek to kanji location.
    info.count  = 0;                                    // Initialize index.
    info.offset = offset;
    for (j = 0; j < stroke_count.size; j++) {           // Count number of kanji.
      if (stroke_count.stroke[j] != i+1) continue;
      info.count++;
      kanji   = (KANJI) j;
      offset += 2;
      fwrite (&kanji,sizeof(kanji),1,file);             //   and write to disk while counting.
    }
    total += info.count;                                // Write index.
    fseek  (file,i*sizeof(radical_info),SEEK_SET);
    fwrite (&info,sizeof(radical_info),1,file);
  }
//
//  Clean up and exit.
//
  fclose (file);
  printf ("done [size=%d kanji=%d]\n",offset,total);
  return (false);
}

//-------------------------------------------------------------------
//
//  Main routine.
//

int main (int argc,char **argv) {
  printf ("RINFO Utility Verison 1.1 copyright (C) 1998,1999 glenn rosenthal\n");
  printf ("  Process Michael Raine & Derc Yamasaki's data files\n\n");
  if (load_file(&stroke_count ,"kanjisrk.dat")) { printf ("Cannot load kanji stroke information\n"); exit (5); }
  if (load_file(&radical_index,"radkanji.idx")) { printf ("Cannot load radical info index\n"      ); exit (5); }
  if (load_file(&radical_data ,"radkanji.dat")) { printf ("Cannot load radical info data\n"       ); exit (5); }
  radical_data .size /= sizeof(KANJI);
  radical_index.size /= sizeof(struct radical_info);
#ifdef READ_STROKE
  if (load_file(&stroke_index ,"stroknji.idx")) { printf ("Cannot load stroke info indxed\n"      ); exit (5); }
  if (load_file(&stroke_data  ,"stroknji.dat")) { printf ("Cannot load stroke info data\n"        ); exit (5); }
  stroke_data  .size /= sizeof(KANJI);
  stroke_index .size /= sizeof(struct radical_info);
  printf ("Stroke indexes=%d  Radical indexes=%d\n",stroke_index.size,radical_index.size);
#else
  printf ("Radical indexes=%d\n",radical_index.size);
#endif
  if (euc_stroke   ("stroke.euc" )) { printf ("Cannot write stroke data in EUC format\n"); exit (5); }
  if (ecu_radical  ("radical.euc")) { printf ("Cannot write radical datain EUC format\n"); exit (5); }
  if (write_stroke ("stroke.dat" )) { printf ("Cannot write kanji stroke information\n" ); exit (5); }
  if (write_radical("radical.dat")) { printf ("Cannot write knaji radical information\n"); exit (5); }
  return (0);
}
